//**********************************//
//RAPPEL DE C++
//**********************************//


//2. Passage d'arguments

//En C et C++, sauf mention explicite les arguments sont passés par copies.
//Ce type de passage est à réserver pour les variables de bases: entiers, float,double
//Il ne faut pas passer de struct, classes, vecteurs par copie: perte d'espace mémoire + perte de temps
//En C, on passe ce type d'arguments par pointeurs.
//En C++, on peut les passer par pointeurs (maClass*) ou référence (maClass&)
//Le pointeur nécessite un déférencement (*mon_objet), alors que la référence permet
// de manipuler l'objet de manière classique.
// Règle standard en C++:
//   * Lorsque l'argument est modifié, on le passe par pointeur.
//   * Lorsque l'argment n'est pas modifié, on le passe par référence constante.
//Modifier un argument passé par référence n'est pas une programmation propre
// et génère des actions inatendues!
//

class mon_vecteur
{

public:
    mon_vecteur(double _x,double _y,double _z);

    //additionne un vecteur à celui courant (notez le passage d'argument en reference constante)
    //Le mot clé const en fin de méthode indique que cette fonction ne modifie pas
    // les paramètres internes (x,y,z) de la classe courante.
    mon_vecteur addition(const mon_vecteur& vec) const;

    //retourne les valeurs de (x,y,z) sans modifier la classe courante
    double get_x()const;
    double get_y()const;
    double get_z()const;

    //permet d'acceder et d'affecter les valeurs de x,y,z car elles sont
    // passés par références.
    // ex. mon_vecteur a(4,5,6);
    //     a.get_x()=78; //on a alors a=(78,5,6);
    // Notez que le mot clé const n'apparait pas en fin de déclaration
    // Le compilateur sait différencier l'appel à get_x()const et get_x() en fonction du contexte.
    double& get_x();
    double& get_y();
    double& get_z();

    // retourne dans les pointeurs les valeurs de x,y,z
    // Notez que l'on prend garde de ne pas écrire sur le pointeur NULL
    void get_values(double *_x,double *_y,double *_z);

private:
    double x,y,z;

};

//implementation
mon_vecteur mon_vecteur::addition(const mon_vecteur& vec) const
{return mon_vecteur(x+vec.get_x(),y+vec.get_y(),z+vec.get_z());}
mon_vecteur::mon_vecteur(double _x,double _y,double _z)
    :x(_x),y(_y),z(_z){}
double mon_vecteur::get_x()const{return x;}
double mon_vecteur::get_y()const{return y;}
double mon_vecteur::get_z()const{return z;}
double& mon_vecteur::get_x() {return x;}
double& mon_vecteur::get_y() {return y;}
double& mon_vecteur::get_z() {return z;}
void mon_vecteur::get_values(double *_x,double *_y,double *_z)
{if(_x!=0) *_x=x; if(_y!=0) *_y=y; if(_z!=0) *_z=z;}



//exemple de classe non correct car défini un comportement non standard
class mon_vecteur2 : public mon_vecteur //par heritage, les méthodes get_x sont accessibles
{
private:
    double x,y,z;
public:
    mon_vecteur2(double _x,double _y,double _z)
            :mon_vecteur(_x,_y,_z){}

    //ici l'appel à addition va modifier le vecteur courant!
    mon_vecteur2 addition(mon_vecteur2& vec)
    {
        vec.get_x()+=x;vec.get_y()+=y;vec.get_z()+=z;
        return vec;
    }
};

//utilisation
int main(int argc,char *argv[])
{
    mon_vecteur a(1,2,3);
    mon_vecteur b(4,5,6);
    mon_vecteur c=a.addition(b); //c=(5,7,9), (a et b) inchangés
    c.get_y()=8; //c=(5,8,9)

    mon_vecteur2 a2(1,2,3);
    mon_vecteur2 b2(4,5,6);
    mon_vecteur2 c2=a2.addition(b2); //c=(5,7,9) mais b2=(5,7,9) aussi!!!

    return 0;
}
