
#include <vector>
#include <list>
#include <set>
#include <map>
#include <string>
#include <iostream>


//**********************************//
//RAPPEL DE C++
//**********************************//


//4. Conteneurs de la STL (Standard Library)

//En C++, on utilisera tant que possible les conteurs de la STL.
//Plus générique, plus sûr que les pointeurs. Plus performant et générique
// que les conteneurs définis manuellement.



class mon_vecteur
{
private:
    double x_interne,y_interne,z_interne;
public:
    mon_vecteur():x_interne(0),y_interne(0),z_interne(0){}
    mon_vecteur(double _x,double _y,double _z):x_interne(_x),y_interne(_y),z_interne(_z){}

    double x() const{return x_interne;}
    double y() const{return y_interne;}
    double z() const{return z_interne;}

    friend std::ostream& operator<<(std::ostream& flux,const mon_vecteur& vec)
    {flux<<"("<<vec.x()<<","<<vec.y()<<","<<vec.z()<<")";return flux;}
};


//classe d'une date de création que l'on utilisera comme clé dans une map
class date
{
   public:
    int jour;
    int mois;
    int annee;


    date():jour(0),mois(0),annee(0){}
    date(int _jour,int _mois,int _annee):jour(_jour),mois(_mois),annee(_annee){}

    friend std::ostream& operator<<(std::ostream& flux,const date& vec)
    {flux<<vec.jour<<"/"<<vec.mois<<"/"<<vec.annee;return flux;}
};

//classement de deux date par un foncteur (classe définissant l'opérateur())
class date_less//classe définissant si date1<date2 ou non par l'appel à l'opérateur parenthèse
{
public:

    //est ce que date1<date2?
    bool operator()(const date& date1,const date& date2)
    {
        if(date1.annee<date2.annee)
            return true;
        if(date1.annee==date2.annee)
        {
            if(date1.mois<date2.mois)
                return true;
            if(date1.mois==date2.mois)
                if(date1.jour<date2.jour)
                    return true;
        }
        return false;
    }
};

//utilisation
int main(int argc,char *argv[])
{


    // ************************** //
    // Vecteur
    // ************************** //

    //un vecteur d'entier
    std::vector<int> vec(10); //idem que: vec int[10];
    //parcours
    for(int k=0,N=vec.size();k<N;++k) //ici N=10 (le conteneur connait sa taille)
        vec[k]=k+8; //acces classique aléatoire comme avec un tableau.

    //par iterateur (commun avec les listes, graphes, ... etc)
    std::vector<int>::iterator it;
    for(it=vec.begin();it!=vec.end();++it)
        *it=vec[1]+1; // it agit comme un pointeur: on peut le déférencer

    //parcours par iterateur constant et remplissage de taille dynamique:
    std::vector<int> vec2;
    for(std::vector<int>::const_iterator it=vec.begin();it!=vec.end();++it)
        vec2.push_back(*it+1); //vec2 augmente sa taille dynamiquement et vaut "vec+1"

    //un vecteur d'un type quelconque
    std::vector<mon_vecteur> V;
    V.push_back(mon_vecteur(4,7,8));
    V.push_back(mon_vecteur(0,1,2));
    //V contient ici 2 classes
    V[1]=V[0];//on affecte directement la valeur de l'entrée 1

    //Dans certains cas, il est nécessaire d'accéder directement aux pointeurs.
    // Le vecteur est stocké de manière contigue, on peut donc revenir à la notation C:
    mon_vecteur* pointeur=&V[0];
    std::cout<<*pointeur<<" ; "<<*(pointeur+1)<<std::endl;
    pointeur[1]=mon_vecteur(-1,-1,-2);

    //A tout moment, on peut dynamiquement changer la taille du vecteur
    V.resize(15);
    V[14]=mon_vecteur(8,8,9);

    //Rappel:
    // Vecteur= accès aléatoire en O(1)
    // Recherche en O(N)
    // Ajout et Suppression d'un élément en milieu de tableau en O(N) (redimentionnement, copie, ...)





    // ************************** //
    // Listes
    // ************************** //

    //les listes permettent des ajouts et des suppressions d'un élément en O(1) peut importe leur positionnement
    //exemple pour une liste de chaine de characteres

    std::list<std::string> ma_liste;
    ma_liste.push_back(" world");
    ma_liste.push_front("hello");
    ma_liste.push_back(" !");

    //affiche "hello world !"
    for(std::list<std::string>::const_iterator it=ma_liste.begin();it!=ma_liste.end();++it)
        std::cout<<*it;
    std::cout<<std::endl;

    //modification d'un mot
    std::list<std::string>::iterator it_modification=ma_liste.begin();
    *it_modification="Good morning"; //modifi le premier mot

    //insere un element en après le mot hello
    std::list<std::string>::iterator it_insertion=ma_liste.begin();
    ++it_insertion; //on se place après "hello"
    ma_liste.insert(it_insertion,std::string(" my very ") + "wonderful"); //concatenation de chaine de caractere

    //suppression de "hello";
    ma_liste.erase(it_insertion); //attention it_insertion est à mettre à jour
    //insertion d'un autre mot
    it_insertion=ma_liste.end();--it_insertion;
    ma_liste.insert(it_insertion," life");

    //affichage de la nouvelle phrase
    for(std::list<std::string>::const_iterator it=ma_liste.begin();it!=ma_liste.end();++it)
        std::cout<<*it;
    std::cout<<std::endl;

    //Rappel:
    // Recherche et accès (aléatoire) en O(N)
    // Ajout et Suppression d'un élément en milieu de tableau en O(1)




    // ************************** //
    // Set est un graphe binaire qui classe automatiquement les éléments du plus petit au plus grand
    // ************************** //

    //exemple d'un arbre d'entier
    std::set<int> mon_arbre;
    mon_arbre.insert(4);
    mon_arbre.insert(7);
    mon_arbre.insert(-8);
    mon_arbre.insert(9);
    mon_arbre.insert(-8);//doublon non ajouté

    //affichage dans l'ordre
    for(std::set<int>::const_iterator it=mon_arbre.begin();it!=mon_arbre.end();++it)
        std::cout<<*it<<" ; ";
    std::cout<<std::endl;

    //suppression d'un élément pointé
    std::set<int>::iterator it_set=mon_arbre.begin();
    ++it_set;++it_set;mon_arbre.erase(it_set);

    //suppression d'un élément donné par la clé (recherche en log(N))
    mon_arbre.erase(4);
    //recherche d'un élément
    std::set<int>::iterator it_set_find=mon_arbre.find(4);//cherche element 4
    if(it_set_find==mon_arbre.end())
        std::cout<<"element 4 non trouve dans le std::set"<<std::endl;

    //affichage dans l'ordre
    for(std::set<int>::const_iterator it=mon_arbre.begin();it!=mon_arbre.end();++it)
        std::cout<<*it<<" ; ";
    std::cout<<std::endl;



    // ************************** //
    // Map est un type particulier de "std::set" contenant une clé pour l'ordonnancement, et un objet associé
    // ************************** //

    //exemple de coordonnées associées à une clée d'ordonnancement (ici une chaine de caractère)
    std::map<std::string,mon_vecteur> ma_map;

    //on insere les element sous forme de paire (clé,objet)
    ma_map.insert(std::make_pair("vecteur_01",mon_vecteur(1,2,3)));
    ma_map.insert(std::make_pair("vecteur_00",mon_vecteur(4,7,8)));
    std::string chaine_1="vecteur_36";mon_vecteur v(4,7,8);
    ma_map.insert(std::make_pair(chaine_1,v));
    ma_map.insert(std::make_pair("vecteur_01",mon_vecteur(-1,-1,-1)));//doublon de clé non ajouté
    ma_map.insert(std::make_pair("vecteur_42",mon_vecteur(7,7,7)));

    //affichage de la map
    std::map<std::string,mon_vecteur>::const_iterator it_map;
    for(it_map=ma_map.begin();it_map!=ma_map.end();++it_map)
        std::cout<<it_map->first<<" : "<<it_map->second<<std::endl;//acces par une paire

    //suppresion d'un element désigné par la clé
    ma_map.erase("vecteur_00");
    //tentative de suppression d'un element non existant
    ma_map.erase("vecteur_420");
    //recherche d'un element
    if(ma_map.find("vecteur_42")!=ma_map.end()) //return .end() si non trouvé
        std::cout<<"element vecteur_42 existe dans la map"<<std::endl;


    //cas d'une clé définie soit même
    // ex. clé = date de création
    std::map<date,std::string,date_less> ma_map2; //notez le 3eme element utilisé pour le classement

    ma_map2.insert(std::make_pair(date(1,1,1889),"construction tour eiffel"));
    ma_map2.insert(std::make_pair(date(18,05,2021),"sortie de Windows 12"));
    ma_map2.insert(std::make_pair(date(29,4,2011),"mariage prince william"));
    ma_map2.insert(std::make_pair(date(17,04,2666),"fin du monde"));
    ma_map2.insert(std::make_pair(date(1,1,-3500),"invention de la roue"));
    ma_map2.insert(std::make_pair(date(15,9,2011),"rentree des classes"));
    ma_map2.insert(std::make_pair(date(14,9,1515),"bataille de Marignan"));

    //suppression d'un element désigné par la clé
    ma_map2.erase(date(15,9,2011));

    //affichage de la map
    for(std::map<date,std::string,date_less>::const_iterator it=ma_map2.begin(),it_end=ma_map2.end();it!=it_end;++it)
        std::cout<<it->first<<" => "<<it->second<<std::endl;



    //Rappel:
    // Recherche et accès (aléatoire) en O(log(N))
    // Ajout et Suppression d'un élément en milieu de tableau en O(log(N))


    return 0;
}
