

#include <iostream>
#include <string>
#include <list>
#include <vector>


//**********************************//
//RAPPEL DE C++
//**********************************//


//5. En C++ il existe 2 approches de généricité
// L'orienté objet par polymorphisme = virtual.
// Les templates.
//
// Avantages/inconvenients principaux:
//  Polymorphisme:
//                 + adaptable dynamiquement
//                 + simple et transparent dans le code
//                 - surchage du temps de calcul à l'execution
//  Template:
//                 + Evalué à la compilation: aucune perte de temps dynamique
//                 - Non dynamique pendant l'execution
//                 - Complexifie le code et la compilation
//


// ********************************************************************** //
//1. Le polymorphisme permet de connaitre le type de classe de manière dynamique.
// ********************************************************************** //

class objet_geometrique
{
public:
    objet_geometrique(const std::string& _nom):nom(_nom){}

    virtual void affiche() const //la méthode virtuelle spécialisée dans chaque cas
    {std::cout<<"Je ne sais pas qui je suis ("<<nom<<")"<<std::endl;}
protected:
    std::string nom;
};

class disque : public objet_geometrique
{
public:
    disque(const std::string& nom,double _centre_x,double _centre_y,double _rayon)
        :objet_geometrique(nom),centre_x(_centre_x),centre_y(_centre_y),rayon(_rayon)
    {}

    virtual void affiche() const
    {
        std::cout<<"Je suis un disque ("<<nom<<") de rayon "<<rayon<<", et de centre ("<<centre_x<<","<<centre_y<<")"<<std::endl;
        //fait son affichage propre...
    }
private:
    double centre_x,centre_y;
    double rayon;
};
class rectangle: public objet_geometrique
{
public:
    rectangle(const std::string& nom,double _x0,double _y0,double _x1,double _y1)
        :objet_geometrique(nom),x0(_x0),y0(_y0),x1(_x1),y1(_y1)
    {}


    virtual void affiche() const
    {
        std::cout<<"Je suis un rectangle ("<<nom<<"), mes cotes sont situes en ("<<x0<<","<<y0<<") ; ("<<x1<<","<<y1<<")"<<std::endl;
        //fait sont affichage propre...
    }
private:
    double x0,y0;
    double x1,y1;
};



// ********************************************************************** //
// 2. Les templates permettent de definir des classes generiques
// ********************************************************************** //


//une seule declaration et implementation de ce vecteur permet de gérer des types divers
template <typename MON_TYPE>
class vecteur_generique
{
private:
    MON_TYPE x,y,z;//ce vecteur contient 3 coordonnees de type quelconque

public:

    vecteur_generique(){}
    //notez que l'on passera par référence constante pour éviter la copie d'élément non connu à l'avance
    vecteur_generique(const MON_TYPE& _x,const MON_TYPE& _y,const MON_TYPE& _z)
        :x(_x),y(_y),z(_z){}

    const MON_TYPE& get_x() const {return x;}
    const MON_TYPE& get_y() const {return y;}
    const MON_TYPE& get_z() const {return z;}

    MON_TYPE& get_x() {return x;}
    MON_TYPE& get_y() {return y;}
    MON_TYPE& get_z() {return z;}

    friend std::ostream& operator<<(std::ostream& flux,const vecteur_generique& vec)
    {
        flux<<vec.get_x()<<","<<vec.get_y()<<","<<vec.get_z();
        return flux;
    }
};

//exemple de méthode templates à plusieurs arguments
//une matrice dont le type est générique et dont la taille est connue à la compilation
template <typename MON_TYPE,int TAILLE_X,int TAILLE_Y>
class matrice_generique
{
private:
    MON_TYPE data[TAILLE_X*TAILLE_Y];
public:

    matrice_generique()
    {
        for(unsigned int k=0;k<TAILLE_X*TAILLE_Y;++k)
            data[k]=0;
    }

    //definition de l'operateur parenthese pour acceder a l'element
    MON_TYPE& operator()(unsigned int kx,unsigned int ky)
    {
        if(kx<TAILLE_X && ky<TAILLE_Y)
            return data[kx+TAILLE_X*ky];
        else
            throw std::exception();
    }

    //dans le cas d'un contexte constant (exemple pour la multiplication)
    const MON_TYPE& operator() (unsigned int kx,unsigned int ky) const
    {
        if(kx<TAILLE_X && ky<TAILLE_Y)
            return data[kx+TAILLE_X*ky];
        else
            throw std::exception();
    }

    //mutiplication entre matrices
    template <int TAILLE_Y2>
    matrice_generique<MON_TYPE,TAILLE_X,TAILLE_Y2> multiply(const matrice_generique<MON_TYPE,TAILLE_Y,TAILLE_Y2>& M)
    {
        matrice_generique<MON_TYPE,TAILLE_X,TAILLE_Y2> matrice_resultat;

        for(unsigned int kx=0;kx<TAILLE_X;++kx)
            for(unsigned int ky=0;ky<TAILLE_Y2;++ky)
                for(unsigned int k=0;k<TAILLE_Y;++k)
                    matrice_resultat(kx,ky) += (*this)(kx,k)*M(k,ky);
        return matrice_resultat;
    }

    friend std::ostream& operator<<(std::ostream& flux,const matrice_generique& M)
    {
        for(unsigned int k=0;k<TAILLE_X;++k)
        {
            flux<<"(";
            for(unsigned int k2=0;k2<TAILLE_Y;++k2)
            {
                flux<<M(k,k2)<<" ";
            }
            flux<<")"<<std::endl;
        }
        return flux;
    }
};

//un exemple de nombre complexe
class nombre_complex
{
private:
    double x;
    double y;
public:

    nombre_complex():x(0.0),y(0.0){}
    nombre_complex(double _x):x(_x),y(0.0){}
    nombre_complex(double _x,double _y):x(_x),y(_y){}

    double get_x() const{return x;}
    double get_y() const{return y;}

    nombre_complex& operator+=(const nombre_complex& z)
    {x+=z.get_x();y+=z.get_y();return *this;}
    nombre_complex& operator*=(const nombre_complex& z)
    {
        double temp_x=x*z.get_x()-y*z.get_y();
        double temp_y=x*z.get_y()+y*z.get_x();
        x=temp_x;y=temp_y;
        return *this;
    }

    friend nombre_complex operator+(const nombre_complex& z1,const nombre_complex& z2)
    {nombre_complex z=z1;z+=z2;return z;}
    friend nombre_complex operator*(const nombre_complex& z1,const nombre_complex& z2)
    {nombre_complex z=z1;z*=z2;return z;}

    friend std::ostream& operator<<(std::ostream& flux,const nombre_complex& z)
    {flux<<z.get_x()<<"+"<<z.get_y()<<"i";return flux;}

};

//utilisation
int main(int argc,char *argv[])
{

    // ************************************** //
    // Polymorphisme par virtual //
    // ************************************** //


    //creation d'objets specifiques
    rectangle *r0=new rectangle("mon rectangle 1",0,0,1,1);
    rectangle *r1=new rectangle("mon rectangle 2",2,2,5,5);
    disque *d0=new disque("mon disque a moi",0,0,1);
    disque *d1=new disque("mon deuxieme disque",5,5,6);


    //creation d'un conteneur generique
    std::list<objet_geometrique*> mon_conteneur;
    mon_conteneur.push_back(r0);
    mon_conteneur.push_back(d0);
    mon_conteneur.push_back(d1);
    mon_conteneur.push_back(r1);

    //affichage spécifique (notez que chaque objet se "souviens" de son type propre)
    for(std::list<objet_geometrique*>::const_iterator it=mon_conteneur.begin();it!=mon_conteneur.end();++it)
        (*it)->affiche();



    //On notera que le polymorphisme nécessite de stocker des pointeurs ou des références.
    //si on avait std::list<objet_geometrique> mon_conteneur, le virtual ne fonctionne plus!
    //ex.
    std::vector<objet_geometrique> mon_conteneur_marche_pas;
    mon_conteneur_marche_pas.push_back(*r0);
    mon_conteneur_marche_pas[0].affiche();//fonctionne pas

    objet_geometrique& ma_reference_a_moi=*r0;
    ma_reference_a_moi.affiche();//fonctionne (par contre on ne peut pas stocker de reference dans un vecteur)

    //on efface la memoire car on a cree des pointeurs (on peut passer par le tableau)
    for(std::list<objet_geometrique*>::const_iterator it=mon_conteneur.begin();it!=mon_conteneur.end();++it)
        delete *it;

    //notons que sans le virtual, on aurait du definir un tableau pour chaque type:
    // std::list<rectangle> liste_de_rectangle;
    // std::list<disque> liste_de_disque;
    // ce qui n'est pas gerable pour un grand nombre d'objets differents.





    // ************************************** //
    // Utilisation de templates
    // ************************************** //

    //divers types géré sans duplication de code
    vecteur_generique<double> mon_vecteur_de_double(4.5,5,8);
    vecteur_generique<float> mon_vecteur_de_float(4.5,5,8);
    vecteur_generique<unsigned int> mon_vecteur_de_uint(4.5,5,8);
    std::cout<<mon_vecteur_de_double<<std::endl;
    std::cout<<mon_vecteur_de_float<<std::endl;
    std::cout<<mon_vecteur_de_uint<<std::endl;

    //mais on peut egalement s'en servir avec ses propres types plus complexes
    vecteur_generique<vecteur_generique<double> > mon_vecteur_de_vecteur;
    mon_vecteur_de_vecteur.get_x()=vecteur_generique<double>(1,5,8);
    mon_vecteur_de_vecteur.get_y()=vecteur_generique<double>(7,8,9);
    mon_vecteur_de_vecteur.get_z()=vecteur_generique<double>(-1.5,0,12.1);
    std::cout<<mon_vecteur_de_vecteur<<std::endl;


    //exemple d'utilisation de la matrice de taille fixée à la compilation
    matrice_generique<nombre_complex,3,2> M1; //matrice de nombre complex de taille (3,2)
    //(on pourrait écrire de manière identique des matrices de doubles ou de int: matrice_generique<double,...> matrice_generique<int,...> ...
    M1(0,0)=nombre_complex(0.5,1);M1(0,1)=nombre_complex(1,1);
    M1(1,0)=nombre_complex(-0.5,0);M1(1,1)=nombre_complex(2,-0.5);
    M1(2,0)=nombre_complex(-0.5,1.5);M1(2,1)=nombre_complex(-0.5,3);

    matrice_generique<nombre_complex,2,3> M2;
    for(unsigned int kx=0;kx<2;++kx)
        for(unsigned int ky=0;ky<3;++ky)
            M2(kx,ky)=nombre_complex(kx+ky+1,kx-ky+2);

    std::cout<<std::endl;
    matrice_generique<nombre_complex,3,3> M=M1.multiply(M2);//multiplication de matrices
    //Notez que si la compilation réussi, on est assuré que les tailles sont compatibles
    // Il n'y a plus besoin de le vérifier pendant l'execution du programme.
    //ex. matrice_generique<double,3,4> A;A*A ne compilera pas!

    std::cout<<M1;
    std::cout<<"*"<<std::endl;
    std::cout<<M2;
    std::cout<<"="<<std::endl;
    std::cout<<M<<std::endl;


    return 0;
}
